use eyre::{Context, Result};
use once_cell::sync::Lazy;
use reqwest::{
    header::{CONTENT_LENGTH, CONTENT_TYPE, DATE},
    Response, StatusCode,
};
use rsa::RsaPrivateKey;
use serde::{de::DeserializeOwned, Deserialize, Serialize};
use tracing::error;

use crate::{
    http::ErrorCode,
    signature::{build_signature, format_date},
};

use super::{SERVER_HEADER, SIGNATURE_HEADER, USER_HEADER};

static APP_USER_AGENT: &str = concat!(env!("CARGO_PKG_NAME"), "/", env!("CARGO_PKG_VERSION"),);

static CLIENT: Lazy<reqwest::Client> = Lazy::new(|| {
    reqwest::Client::builder()
        .user_agent(APP_USER_AGENT)
        .build()
        .unwrap()
});

pub fn is_valid_domain(domain: &str) -> bool {
    addr::parse_domain_name(domain).is_ok()
}

pub async fn get<R: DeserializeOwned>(
    private_key: &RsaPrivateKey,
    self_domain: &str,
    host: &str,
    path: &str,
    user_id: Option<String>,
) -> Result<Option<R>> {
    let (signature, date) = build_signature(private_key, host, path, None, user_id.clone());

    let mut req = CLIENT
        .get(format!("https://{host}{path}"))
        .header(DATE, format_date(date))
        .header(SIGNATURE_HEADER, signature)
        .header(SERVER_HEADER, self_domain);

    req = if let Some(id) = user_id {
        req.header(USER_HEADER, id)
    } else {
        req
    };

    let resp = req.send().await?;
    handle_response(resp).await.wrap_err("handling response")
}

pub async fn post<T: Serialize, R: DeserializeOwned>(
    private_key: &RsaPrivateKey,
    self_domain: &str,
    host: &str,
    path: &str,
    user_id: Option<String>,
    body: &T,
) -> Result<Option<R>> {
    let body = serde_json::to_string(body)?;

    let (signature, date) =
        build_signature(private_key, host, path, Some(body.len()), user_id.clone());

    let mut req = CLIENT
        .post(format!("https://{host}{path}"))
        .header(DATE, format_date(date))
        .header(SIGNATURE_HEADER, signature)
        .header(SERVER_HEADER, self_domain)
        .header(CONTENT_TYPE, "application/json; charset=utf-8")
        .header(CONTENT_LENGTH, body.len().to_string())
        .body(body);

    req = if let Some(id) = user_id {
        req.header(USER_HEADER, id)
    } else {
        req
    };

    let resp = req.send().await?;
    handle_response(resp).await.wrap_err("handling response")
}

async fn handle_response<R: DeserializeOwned>(resp: Response) -> Result<Option<R>, ResponseError> {
    if resp.status() == StatusCode::NO_CONTENT {
        return Ok(None);
    }

    if resp.status() != StatusCode::OK {
        let status = resp.status().as_u16();

        let error_json = resp
            .json::<ApiError>()
            .await
            .map_err(|_| ResponseError::JsonError)?;

        return Err(ResponseError::NotOk {
            status,
            code: error_json.code,
            message: error_json.message,
        });
    }

    let parsed = resp
        .json::<R>()
        .await
        .map_err(|_| ResponseError::JsonError)?;
    Ok(Some(parsed))
}

#[derive(thiserror::Error, Debug, Clone)]
pub enum ResponseError {
    #[error("non-200 status code")]
    NotOk {
        status: u16,
        code: ErrorCode,
        message: String,
    },
    #[error("error deserializing JSON")]
    JsonError,
}

#[derive(Deserialize)]
struct ApiError {
    pub code: ErrorCode,
    pub message: String,
}
