use std::num::ParseIntError;

use axum::{
    http::{StatusCode, header::ToStrError},
    response::{IntoResponse, Response},
    Json,
};
use eyre::Report;
use serde::{Serialize, Deserialize};
use serde_json::json;
use tracing::error;

use crate::FoxError;

pub struct ApiError {
    pub status: StatusCode,
    pub code: ErrorCode,
    pub message: String,
}

impl IntoResponse for ApiError {
    fn into_response(self) -> Response {
        (
            self.status,
            Json(json!({
                "status": self.status.as_u16(),
                "code": self.code,
                "message": self.message,
            })),
        )
            .into_response()
    }
}

#[derive(Serialize, Deserialize, Debug, Clone)]
#[serde(rename_all = "SCREAMING_SNAKE_CASE")]
pub enum ErrorCode {
    InternalServerError,
    ObjectNotFound,
    InvalidServer,
    InvalidHeader,
    InvalidDate,
    InvalidSignature,
    MissingSignature,
    GuildNotFound,
    Unauthorized,
    InvalidRestEvent,
}

impl From<sqlx::Error> for ApiError {
    fn from(err: sqlx::Error) -> Self {
        ApiError {
            status: StatusCode::INTERNAL_SERVER_ERROR,
            code: ErrorCode::InternalServerError,
            message: if cfg!(debug_assertions) {
                format!("Database error: {}", err)
            } else {
                "Database error".into()
            },
        }
    }
}

impl From<Report> for ApiError {
    fn from(err: Report) -> Self {
        match err.downcast_ref::<FoxError>() {
            Some(e) => return (*e).into(),
            None => {}
        };

        error!("Error in handler: {}", err);
        ApiError {
            status: StatusCode::INTERNAL_SERVER_ERROR,
            code: ErrorCode::InternalServerError,
            message: if cfg!(debug_assertions) {
                format!("Internal server error: {}", err)
            } else {
                "Internal server error".into()
            },
        }
    }
}

impl From<FoxError> for ApiError {
    fn from(err: FoxError) -> Self {
        match err {
            FoxError::NotFound => ApiError {
                status: StatusCode::NOT_FOUND,
                code: ErrorCode::ObjectNotFound,
                message: "Object not found".into(),
            },
            FoxError::DatabaseError => ApiError {
                status: StatusCode::INTERNAL_SERVER_ERROR,
                code: ErrorCode::InternalServerError,
                message: "Database error".into(),
            },
            FoxError::SignatureDateOutOfRange(s) => ApiError {
                status: StatusCode::BAD_REQUEST,
                code: ErrorCode::InvalidSignature,
                message: format!("Signature date out of range: {}", s),
            },
            FoxError::ResponseNotOk => ApiError {
                status: StatusCode::INTERNAL_SERVER_ERROR,
                code: ErrorCode::InternalServerError,
                message: "Error response from remote server".into(),
            },
            FoxError::InvalidServer => ApiError {
                status: StatusCode::BAD_REQUEST,
                code: ErrorCode::InvalidServer,
                message: "Invalid remote server".into(),
            },
            FoxError::InvalidHeader => ApiError {
                status: StatusCode::BAD_REQUEST,
                code: ErrorCode::InvalidHeader,
                message: "Invalid header value".into(),
            },
            FoxError::InvalidDate => ApiError {
                status: StatusCode::BAD_REQUEST,
                code: ErrorCode::InvalidHeader,
                message: "Invalid date value in header".into(),
            },
            FoxError::MissingSignature => ApiError {
                status: StatusCode::BAD_REQUEST,
                code: ErrorCode::MissingSignature,
                message: "Missing signature".into(),
            },
            FoxError::InvalidSignature => ApiError {
                status: StatusCode::BAD_REQUEST,
                code: ErrorCode::InvalidSignature,
                message: "Invalid signature".into(),
            },
            FoxError::MissingUser => ApiError {
                status: StatusCode::BAD_REQUEST,
                code: ErrorCode::InvalidHeader,
                message: "Missing user header".into(),
            },
            FoxError::NotInGuild => ApiError {
                status: StatusCode::NOT_FOUND,
                code: ErrorCode::GuildNotFound,
                message: "Channel or guild not found".into(),
            },
            FoxError::Unauthorized => ApiError {
                status: StatusCode::UNAUTHORIZED,
                code: ErrorCode::Unauthorized,
                message: "Missing or invalid token".into(),
            },
            FoxError::ChannelNotFound => ApiError {
                status: StatusCode::NOT_FOUND,
                code: ErrorCode::GuildNotFound,
                message: "Channel or guild not found".into(),
            },
            FoxError::ProxyInternalServerError => ApiError {
                status: StatusCode::INTERNAL_SERVER_ERROR,
                code: ErrorCode::InternalServerError,
                message: "Internal server error".into(),
            },
            FoxError::InvalidRestEvent => ApiError {
                status: StatusCode::BAD_REQUEST,
                code: ErrorCode::InvalidRestEvent,
                message: "Invalid RestEvent for this instance type".into(),
            },
        }
    }
}

impl From<ToStrError> for ApiError {
    fn from(_: ToStrError) -> Self {
        ApiError {
            status: StatusCode::BAD_REQUEST,
            code: ErrorCode::InvalidHeader,
            message: "Invalid header value".into(),
        }
    }
}

impl From<chrono::ParseError> for ApiError {
    fn from(_: chrono::ParseError) -> Self {
        ApiError {
            status: StatusCode::BAD_REQUEST,
            code: ErrorCode::InvalidDate,
            message: "Invalid date header value".into(),
        }
    }
}

impl From<ParseIntError> for ApiError {
    fn from(_: ParseIntError) -> Self {
        ApiError {
            status: StatusCode::BAD_REQUEST,
            code: ErrorCode::InvalidHeader,
            message: "Invalid content length value".into(),
        }
    }
}
